#!/usr/bin/env python3

import hashlib
import re
import xml.etree.ElementTree as etree  # noqa: N813
from typing import Any
from urllib.parse import unquote as urlunquote
from urllib.parse import quote as urlquote
from urllib.parse import urlparse

import markdown
from markdown.extensions import Extension
from markdown.extensions.footnotes import (FootnoteExtension, FootnoteBlockProcessor, FootnoteInlineProcessor,
                                           FootnoteTreeprocessor, FootnotePostTreeprocessor, FootnotePostprocessor)
from markdown.inlinepatterns import InlineProcessor, SimpleTagInlineProcessor
from markdown.preprocessors import Preprocessor
from markdown.util import STX, ETX

from .config import Context

PRIO_BASE = 10000  # priority base for our extensions

# Patterns for ObsidianImages
OBSIMAGE_PATTERN = r'!\[\[(.*?)\]\]'
GENERICIMAGE_PATTERN = r'!\[(.*?)\]\((.*?)\)'

# Patterns for ObsidianLinks
OBSLINK_PATTERN = r'\[\[(.*?)\]\]'
GENERICLINK_PATTERN = r'\[(.*?)\]\((.*?)\)'

INLINE_FOOTNOTE_REF_PREFIX = STX + "erbosoft_fnref:"
INLINE_FOOTNOTE_REF_PATTERN = INLINE_FOOTNOTE_REF_PREFIX + "{}" + ETX

def is_proper_url(s: str) -> bool:
    """
    Checks to see if a string is a "proper" URL.

    Args:
        s (str): The string to be parsed.

    Returns:
        bool: ``True`` if the string is a proper URL, ``False`` if not.
    """
    parseout = urlparse(s)
    return True if parseout.scheme else False


def sanitize_reference(s: str) -> str:
    """
    Sanitizes an internal reference to a file by removing URL-quoted characters and any Markdown suffix.

    Args:
        s (str): Th reference to be sanitized.

    Returns:
        str: The sanitized reference.
    """
    rc = urlunquote(s)
    if rc.endswith('.md'):
        rc = rc[:-3]
    return rc


def find_extension(md: markdown.Markdown, cls: type[Extension]) -> Extension | None:
    """
    Locate a registered extension in the Markdown parser.

    Args:
        md (markdown.Markdown): The Markdown parser to look through.
        cls (type): The class of the extension to be retrieved.

    Returns:
        Extension: The retrieved extension, or ``None`` if it was not found.
    """
    for ex in md.registeredExtensions:
        if isinstance(ex, cls):
            return ex
    return None


class MetaStripper(Extension):
    """
    An extension that strips the metadata off the front of Obsidian pages, as it's already been parsed in an
    earlier step.
    """

    class MetaStripperProc(Preprocessor):
        """The actual metadata stripper."""
        def run(self, lines: list[str]) -> list[str]:
            """
            Removes the metadata lines from the front of the array of lines.

            Args:
                lines (list[str]): The Markdown lies read in.

            Returns:
                list[str]: The Markdown lines after the metadata has been removed.
            """
            if lines[0] == '---':
                lines.pop(0)
                while lines[0] != '---':
                    lines.pop(0)
                lines.pop(0)
            return lines

    def extendMarkdown(self, md: markdown.Markdown) -> None:  # noqa: N802
        """
        Registers the metadata stripper with the Markdown parser.

        Args:
            md (markdown.Markdown): The Markdown parser to register the stripper with.
        """
        md.preprocessors.register(MetaStripper.MetaStripperProc(md), 'metastripper', PRIO_BASE)


class ObsidianImages(Extension):
    """An extension that supports image tags the way Obsidian handles them."""
    DIMS = re.compile(r'(.*)\|(\d+)(?:x(\d+))?')

    def __init__(self, context: Context, **kwargs: dict[str, Any]) -> None:
        """
        Initialize the image processor class.

        Args:
            context (Context): The dragonglass context, containing the configuration.
            **kwargs (dict): Other keyword arguments to the image tag processor.
        """
        super(ObsidianImages, self).__init__(**kwargs)
        self._context = context

    @property
    def invalid_reference_classname(self) -> str:
        """
        Returns the CSS class name for an invalid reference in the text.
        """
        return 'invalid-reference'

    def _parse_dimensions(self, s: str) -> tuple[str, int, int]:
        """
        Parses an image reference looking for image dimensions.

        Args:
            s (str): The image reference to be parsed.

        Returns:
            str: The actual image reference.
            int: The image width in pixels, or -1 if not specified.
            int: The image height in pixels, or -1 if not specified.
        """
        m = self.DIMS.match(s)
        if m:
            width = int(m.group(2))
            height = int(m.group(3)) if m.group(3) else -1
            return m.group(1), width, height
        else:
            return s, -1, -1

    def _lookup_image_reference(self, name: str) -> str | None:
        assert self._context.src_index is not None
        node, _ = self._context.src_index.lookup(name)
        if node:
            return node.link_target(self._context.url_prefix)
        return None

    class ObsidianImageProc(InlineProcessor):
        def __init__(self, pattern: str, md: markdown.Markdown, extref: Any) -> None:
            super(ObsidianImages.ObsidianImageProc, self).__init__(pattern, md)
            self._extref = extref

        def handleMatch(self, m: re.Match[str], data: str) -> tuple[etree.Element, int, int]:  # noqa: N802
            name, width, height = self._extref._parse_dimensions(m.group(1))
            link = self._extref._lookup_image_reference(name)
            if link is None:
                el = etree.Element('span')
                el.set('class', self._extref.invalid_reference_classname)
                el.text = name
            else:
                el = etree.Element('img')
                el.set('src', link)
                el.set('alt', name)
                if width > 0:
                    el.set('width', str(width))
                if height > 0:
                    el.set('height', str(height))
            return el, m.start(0), m.end(0)

    class GenericImageProc(InlineProcessor):
        def __init__(self, pattern: str, md: markdown.Markdown, extref: Any) -> None:
            super(ObsidianImages.GenericImageProc, self).__init__(pattern, md)
            self._extref = extref

        def handleMatch(self, m: re.Match[str], data: str) -> tuple[etree.Element, int, int]:  # noqa: N802
            name, width, height = self._extref._parse_dimensions(m.group(1))
            link = m.group(2)
            if is_proper_url(link):
                el = etree.Element('img')
                el.set('src', link)
                if len(name) > 0:
                    el.set('alt', name)
                if width > 0:
                    el.set('width', str(width))
                if height > 0:
                    el.set('height', str(height))
            else:
                newlink = self._extref._lookup_image_reference(link)
                if newlink:
                    el = etree.Element('img')
                    el.set('src', newlink)
                    if len(name) > 0:
                        el.set('alt', name)
                    if width > 0:
                        el.set('width', str(width))
                    if height > 0:
                        el.set('height', str(height))
                else:
                    el = etree.Element('span')
                    el.set('class', self._extref.invalid_reference_classname)
                    el.text = link
            return el, m.start(0), m.end(0)

    def extendMarkdown(self, md: markdown.Markdown) -> None:  # noqa: N802
        md.inlinePatterns.register(ObsidianImages.ObsidianImageProc(OBSIMAGE_PATTERN, md, self),
                                   'obsidian_images', PRIO_BASE + 1010)
        md.inlinePatterns.register(ObsidianImages.GenericImageProc(GENERICIMAGE_PATTERN, md, self),
                                   'obsidian_generic_images', PRIO_BASE + 1000)


class ObsidianLinks(Extension):
    """
    An extension that processes Obsidian internal links in the [[page name]] format, as well as overrides the standard
    Markdown link processing to handle Obsidian internal links as well as external links.
    """
    def __init__(self, context: Context, **kwargs: dict[str, Any]) -> None:
        super(ObsidianLinks, self).__init__(**kwargs)
        self._context = context

    @property
    def obsidian_link_classname(self) -> str:
        return 'obsidian-link'

    @property
    def invalid_reference_classname(self) -> str:
        return 'invalid-reference'

    def _parse_reference(self, contents: str) -> tuple[str | None, str]:
        text = None
        t = contents.split('|')
        if len(t) > 1:
            text = t[1]
            contents = t[0]

        assert self._context.src_index is not None
        node, _ = self._context.src_index.lookup(contents)
        if not text:
            text = contents
        if node:
            return node.link_target(self._context.url_prefix), text
        return None, text

    class ObsidianLinksProc(InlineProcessor):
        def __init__(self, pattern: str, md: markdown.Markdown, extref: Any) -> None:
            super(ObsidianLinks.ObsidianLinksProc, self).__init__(pattern, md)
            self._extref = extref

        def handleMatch(self, m: re.Match[str], data: str) -> tuple[etree.Element, int, int]:  # noqa: N802
            link, text = self._extref._parse_reference(m.group(1))
            if link is None:
                el = etree.Element('span')
                el.set('class', self._extref.invalid_reference_classname)
                el.text = text
            else:
                el = etree.Element('a')
                el.set('href', link)
                el.set('class', self._extref.obsidian_link_classname)
                el.text = text
            return el, m.start(0), m.end(0)

    class GenericLinksProc(InlineProcessor):
        def __init__(self, pattern: str, md: markdown.Markdown, extref: Any) -> None:
            super(ObsidianLinks.GenericLinksProc, self).__init__(pattern, md)
            self._extref = extref

        def handleMatch(self, m: re.Match[str], data: str) -> tuple[etree.Element, int, int]:  # noqa: N802
            text = m.group(1)
            link = m.group(2)
            if link.startswith('<') and link.endswith('>'):  # handle whitespace encoding
                link = urlquote(link[1:-1].strip(), safe=':/')
            if is_proper_url(link):
                el = etree.Element('a')
                el.set('href', link)
                el.text = text
            else:
                newlink, _ = self._extref._parse_reference(sanitize_reference(link))
                if newlink is None:
                    el = etree.Element('span')
                    el.set('class', self._extref.invalid_reference_classname)
                    el.text = text
                else:
                    el = etree.Element('a')
                    el.set('href', newlink)
                    el.set('class', self._extref.obsidian_link_classname)
                    el.text = text
            return el, m.start(0), m.end(0)

    def extendMarkdown(self, md: markdown.Markdown) -> None:  # noqa: N802
        md.inlinePatterns.register(ObsidianLinks.ObsidianLinksProc(OBSLINK_PATTERN, md, self),
                                   'obsidian_links', PRIO_BASE + 110)
        md.inlinePatterns.register(ObsidianLinks.GenericLinksProc(GENERICLINK_PATTERN, md, self),
                                   'obsidian_generic_links', PRIO_BASE + 100)


class ObsidianInlines(Extension):
    """An extension that handles the special Obsidian markdown format sequences."""
    def extendMarkdown(self, md: markdown.Markdown) -> None:  # noqa: N802
        """
        Registers the inline patterns with the Markdown parser.

        Args:
            md (markdown.Markdown): The Markdown parser to register the patterns with.
        """
        md.inlinePatterns.register(SimpleTagInlineProcessor(r'()~~(.*?)~~', 'del'), 'strikeout', PRIO_BASE)
        md.inlinePatterns.register(SimpleTagInlineProcessor(r'()\=\=(.*?)\=\=', 'ins'), 'highlight', PRIO_BASE + 1)


class ObsidianStyleFootnotes(FootnoteExtension):
    """
    An extension of the standard Markdown processor footnote extension to support Obsidian-style inline footnotes
    in addition to standard ones.
    """
    class BlockProcessor(FootnoteBlockProcessor):
        """A block processor extended to handle Obsidian inline footnotes."""
        RE_INLINE = re.compile(r'\^\[(.+)]')

        def run(self, parent: etree.Element, blocks: list[str]) -> bool:
            """ Find, set, and remove footnote definitions. """
            block = blocks.pop(0)
            m_block = self.RE.search(block)
            m_inline = self.RE_INLINE.search(block)
            if m_block or m_inline:

                if m_block is not None and m_inline is not None:
                    if m_block.start() > m_inline.start():
                        m_block = None  # inline goes first

                if m_block:
                    id = m_block.group(1)
                    fn_blocks = [m_block.group(2)]

                    # Handle rest of block
                    therest = block[m_block.end():].lstrip('\n')
                    m2 = self.RE.search(therest)
                    if m2:
                        # Another footnote exists in the rest of this block.
                        # Any content before match is continuation of this footnote, which may be lazily indented.
                        before = therest[:m2.start()].rstrip('\n')
                        fn_blocks[0] = '\n'.join([fn_blocks[0], self.detab(before)]).lstrip('\n')
                        # Add back to blocks everything from beginning of match forward for next iteration.
                        blocks.insert(0, therest[m2.start():])
                    else:
                        # All remaining lines of block are continuation of this footnote, which may be lazily indented.
                        fn_blocks[0] = '\n'.join([fn_blocks[0], self.detab(therest)]).strip('\n')

                        # Check for child elements in remaining blocks.
                        fn_blocks.extend(self.detectTabbed(blocks))

                    footnote = "\n\n".join(fn_blocks)
                    self.footnotes.setFootnote(id, footnote.rstrip())

                    if block[:m_block.start()].strip():
                        # Add any content before match back to blocks as separate block
                        blocks.insert(0, block[:m_block.start()].rstrip('\n'))
                    return True

                if m_inline:
                    footnote = m_inline.group(1).strip()
                    id = hashlib.sha1(bytes(footnote, 'utf-8')).hexdigest()
                    self.footnotes.setFootnote(id, footnote)
                    blocks.insert(0, block[:m_inline.start(0)].rstrip()
                                  + INLINE_FOOTNOTE_REF_PATTERN.format(id) + block[m_inline.end(0):])
                    return True

            # No match. Restore block.
            blocks.insert(0, block)
            return False

    class InlineReplacer(InlineProcessor):
        def __init__(self, pattern: str, extref: Any) -> None:
            super(ObsidianStyleFootnotes.InlineReplacer, self).__init__(pattern)
            self._extref = extref

        def handleMatch(self, m: re.Match[str], data: str) -> tuple[etree.Element, int, int]:  # noqa: N802
            etxpoint = data.find(ETX, m.end(0))
            assert etxpoint > m.end(0)
            id = data[m.end():etxpoint]
            assert id in self._extref.footnotes.keys()
            sup = etree.Element('sup')
            a = etree.SubElement(sup, 'a')
            sup.set('id', self._extref.makeFootnoteRefId(id, found=True))
            a.set('href', '#' + self._extref.makeFootnoteId(id))
            a.set('class', 'footnote-ref')
            a.text = self._extref.getConfig('SUPERSCRIPT_TEXT').format(
                list(self._extref.footnotes.keys()).index(id) + 1
            )
            return sup, m.start(0), etxpoint + 1

    def extendMarkdown(self, md):
        """
        Registers the footnote processor with the Markdown parser.

        Args:
            md (markdown.Markdown): The Markdown parser to register the footnote processor with.
        """
        md.registerExtension(self)
        self.parser = md.parser
        self.md = md
        # Insert a `blockprocessor` before `ReferencePreprocessor`
        md.parser.blockprocessors.register(ObsidianStyleFootnotes.BlockProcessor(self), 'footnote', 17)

        # Insert an inline pattern before `ImageReferencePattern`
        FOOTNOTE_RE = r'\[\^([^\]]*)\]'  # blah blah [^1] blah
        md.inlinePatterns.register(FootnoteInlineProcessor(FOOTNOTE_RE, self), 'footnote', 175)
        md.inlinePatterns.register(ObsidianStyleFootnotes.InlineReplacer(INLINE_FOOTNOTE_REF_PREFIX, self),
                                   'footnote2', 175)
        # Insert a tree-processor that would actually add the footnote div
        # This must be before all other tree-processors (i.e., `inline` and
        # `codehilite`) so they can run on the the contents of the div.
        md.treeprocessors.register(FootnoteTreeprocessor(self), 'footnote', 50)

        # Insert a tree-processor that will run after inline is done.
        # In this tree-processor we want to check our duplicate footnote tracker
        # And add additional `backrefs` to the footnote pointing back to the
        # duplicated references.
        md.treeprocessors.register(FootnotePostTreeprocessor(self), 'footnote-duplicate', 15)

        # Insert a postprocessor after amp_substitute processor
        md.postprocessors.register(FootnotePostprocessor(self), 'footnote', 25)


def create_markdown_parser(context: Context) -> markdown.Markdown:
    """
    Creates a Markdown parser with all our extensions loaded.

    Args:
        context (Context): The dragonglass context, that contains the configuration.

    Returns:
        markdown.Markdown: The new Markdown parser.
    """
    return markdown.Markdown(extensions=[MetaStripper(),
                                         ObsidianStyleFootnotes(SUPERSCRIPT_TEXT='[{}]', SEPARATOR='-'),
                                         ObsidianImages(context),
                                         ObsidianLinks(context),
                                         ObsidianInlines()])
